#!/usr/bin/python

#from nx5d.xrd import data as xd
#from nx5d.xrd import kmc3 as kmc3

import pytest, random

from pprint import pprint

@pytest.fixture
def testfile():

    # Note that, for data storage reasons within the git project,
    # our testfile contains only TIFF data for scan "2.1".
    return "./tests/test_data/spech5/231-cw7-12083-roessle.spec"


#def test_experimentsetup(testfile):
#
#    h5like = kmc3.SpecTiffH5(testfile)
#    
#    templ = kmc3.ExperimentTemplate
#    
#    setupGen = xd.SetupGenerator(templ)
#    setup = setupGen(h5like)
#
#    for i in templ:
#        assert i in setup
#
#    for i in setup:
#        assert i in templ

#    # Test setting of values fpr keys that already exist.
#    ev = random.random()*1000
#    assert xd.SetupGenerator(templ, beamEnergy=ev)(h5like)['beamEnergy'] == ev
#
#    # Test that keys not in template are rejected
#    with pytest.raises(KeyError):
#        xd.SetupGenerator(templ, jinkies=ev)
#    
#    #print("Experiment template:")
#    #pprint(templ)
#    #print("Experiment definition:")
#    #pprint(setup)



#!/usr/bin/python3

from silx.io.spech5 import SpecH5, SpecH5LazyNodeDataset, SpecH5NodeDataset
from silx.io.commonh5 import Group as H5Group, SoftLink as H5SoftLink
from h5py import File as H5File
from os import path
import tifffile
import numpy as np

import logging

logger = logging.getLogger(__name__)

'''
Provides a series of objects to mock an "ESRF dialect" of a HDF5-based Nexus
data file for X-ray diffraction experiments for SPEC files.
As an extra provision, detector data is loaded from external `.tiff` files
with an easily computable path. This is typical for older installation of
SPEC which couldn't handle and store large(-ish) 2D detector images.

The "ESRF dialect" is essentially this:

  - There is 1D or 2D detector data ("images" or "frames").

  - The data is organized in "scans", i.e. images are oranized in series.
    One scan typically consists of several images with the same hard characteristics
    (i.e. same geometry), but taken under systematically differing external parameters:
    e.g. by "scanning" one or several parameter devices (angle? time delay?) and
    obtaining an image at each of the scan points.

  - The data format where the images are stored is a 3-dimensional block of NxWxH,
    where 'WxH' is the width/height of one frame, and 'N' is the number of frames
    in a scan. For a scan performed along a single paramter, each subsequent frame
    corresponds to a different parameter instance; for a scan taken along 2 parameters
    (e.g. inside a 2D parameter space), the scans are interleaved: there are actually
    (n*m)xWxH frames. This is still a 3D dataset with images of geometry WxH, but
    the first "chunk" of 'n' images are scanned along the first paramter and a specific
    'm' value (i.e. all have the same 'm'), the next chunk is scanned again along the
    whole N space but for a different 'm', and so on.

  - Besides detector data, there are also associated metadata -- e.g. the parameter
    arrays along which the scans took place. These are also Nx(...) data sets; in the
    simplest cases, they are 2D (parameter is scalar).

  - The both detector data and scan metadata are stored at specific "paths" within
    the data container. In HDF5 terminology, this is the literal HDF5 path within
    the file (e.g. "data/rayonix/detector"). We don't make any assumptions about
    the tree organisation (yes, typically it follows NX data format rules, but...).

  - A data container (i.e. a "file") can contain one or several scans. Each scan
    has a top-level subdirectory of its own.

  - Beyond data series (detector and metadata), there can also exist single scalar
    values setting the state for the entire experiment (e.g. "photon_energy", or
    "center_pixel", ...).

  - [FIXME: haven't decided yet whether top-level scalar metadata still needs to
    be embedded into a scan (i.e. is per-scan only), or can also be at the very
    top. I think ESRF has the former...]

'''

class SpecH5LazyTiffNode(SpecH5LazyNodeDataset):
    '''
    Implements a NodeDataset that lazily loads a list of TIFF files,
    specified as a path format that relies on an index variable.
    '''
    def __init__(self, *args, tiffPathFmt="{frameidx}.tiff", numFrames=0, cache=True, **kwargs):
        super().__init__(*args, **kwargs)
        self._tiff_path_fmt = tiffPathFmt
        self._tiff_frames_no = int(numFrames)
        self._cache_tiffs = cache


    def _create_data(self):
        data = []
        shape = None
        lasterr = None
        missing = []
        for i in range(self._tiff_frames_no):
            try:
                tiff = tifffile.imread(self._tiff_path_fmt.format(frameidx=i))
                shape = tuple([s for s in tiff.shape])
            except FileNotFoundError as e:
                lasterr = e
                if len(missing) == 0:
                    logger.warning(str(e)+". Assuming bad frame, "
                                    "similar errors will be suppressed.")
                missing.append(i)
                tiff = None
            data.append(tiff)


        # It happens a lot that individual files are missing, so
        # we're just ignoring those and replace them with NaN arrays
        # of the appropriate shape. But if we have no shape at this
        # point, this means that _all_ TIFFs are missing. We
        # want to get loud about that (don't we?...)
        if shape is None:
            logger.warning("Dataset had no frames at all.")
            return np.array(np.nan)
        
        a = np.array([ (i if i is not None else np.full(shape, np.nan)) for i in data])
        
        return a


    def _get_data(self):
        if self._cache_tiffs:
            return super()._get_data()

        # Avoid caching
        return self._create_data()


    def __getitem__(self, item):
        if not self._cache_tiffs:
            data = self._create_data()
            return data.__getitem__(item)
        return super().__getitem__(item)

    
class SpecTiffH5(SpecH5):
    '''
    Mocks a HDF5 file using the Spec H5 module from Silx.

    In addition the "regular" `slix.io.spech5.SpecH5` functionality, after loading the
    SPEC file, this module also injects lazy-loading nodes for the TIFF files
    correspondng to a typical KMC3 setup.
    '''

    def __init__(self, *args, instrumentName="pilatus",
                 framePathFmt='{dirname}/{instr}/S{scannr:05}/{basename}_{scannr}_{frameidx}.tif',
                 cache=True,
                 **kwargs):
        ''' Initializes a HDF5-like object based on a SPEC file with data in TIFF images.

        Args:
            instrumentName: A name/label for the instrument. For one, this is used to compile
              the location of where to insert the data from the external TIFF files (as
              specified by `tiffInsertFmt` and `tiffLinkFmt`); for another, this is used
              to determine the location of the TIFF files on disk relative to the
              SPEC file (as specified by `scanPathFmt`).
        
            framePathFmt: format for the image scan filepath.

            scanIndexNameFmt: string formatting that ties the scan number (on disk)
              and scan name (in teh SpecH5 naming) together.

            cache: This parameter is passed to `SpecH5LazyNode`. If set to `True` (the default),
              data from TIFF files in subnodes is cached in memory. This is generally what you
              want, except for rare cases where you are going to read all the data in memory
              first, and risk overwhelming your local RAM. This is the case, for instance,
              when you call `silx.io.convert.write_to_h5()` on a `SpecTiffH5` object which
              represents a large set of scans.

         All string formatters can make use of any or all of the following variables:
        
           - `{dirname}`: folder-only component of the spec file name

           - `{instr}`: the instrument name (from the `instrumentName` parameter above)
        
           - `{basename}`: base name the SPEC file, with the last component after a dot
             (typically ".spec" removed

           - `{scannr}`: scan number; this is typically an integer, assigned by "the
             algorithm", and typically starting with 1 for the first scan

           - `{scanidx}`: scan index; this is an integer starting with 0 and designating
             the cosecutive index value of the scan.

           - `{frameidx}`: the index (starting with 0) of the frame being processed
             (i.e. the TIFF file).

        In addition to the explicit parameters, we also use the following positional
        parameters from `args`:

           - `filename` (position 0): the path of the SPEC file to read
        
        '''
        super().__init__(*args, **kwargs)
        
        self._specBaseName = '.'.join(path.basename(self.filename).split('.')[:-1]) \
            or path.base(self.filename)
        
        self._specDirName = path.abspath(path.dirname(self.filename))

        for scanName in self.keys():
            
            scan = self[scanName]
            scanNr, foo = scanName.split('.')

            tiffPartialFmt = framePathFmt.format(**{
                "basename": self._specBaseName,
                "dirname": self._specDirName,
                "scannr": int(scanNr),
                "instr": instrumentName,
                "scanidx": "{scanidx}",
                "frameidx": "{frameidx}"
            })

            datakw = {'cache': cache}
            self._adopt_scan(scan, instrumentName, tiffPartialFmt, datakw)

            self._enter_cnt = 0
            self._enter_data = None


    def __enter__(self, *args, **kwargs):
        '''
        Opening / closing a SpecTiffH5 is pretty expensive. We usually want to
        employ with-guards (enter/exit) when doing that, but the original SpecH5
        does not support "nested" guards (i.e. entering/exitting multiple
        times on the same object).

        This is an extension to do just that. (Need to check if h5py.File supports
        this... the original Python open() / file object definitely does.)
        '''
        if self._enter_cnt == 0:
            self._enter_data = super().__enter__(*args, **kwargs)
        self._enter_cnt += 1
        return self._enter_data


    def __exit__(self, *args, **kwargs):
        if self._enter_cnt > 0:
            self._enter_cnt -= 1
        if self._enter_cnt == 0:
            super().__exit__(*args, **kwargs)


    def _adopt_scan(self, scanObj, instrumentName, tiffFormat, dataNodeArgs=None):

        measRoot = scanObj['measurement']
        instrRoot = scanObj['instrument']
        posnrs = instrRoot['positioners']
        

        # Let's hope they're all the same... :-)
        numFrames = 0
        #print("Positioners:", posnrs.keys())
        for p in posnrs.keys():
            #print("Positioner:", p, "length", posnrs[p].size)
            numFrames = max(numFrames, posnrs[p].size)

        # place data inside "instrument/<name>/data"
        instrRoot.add_node(H5Group(parent=instrRoot, name=instrumentName,
                                   attrs={'NX_class': 'NXinstrument'} ))
        instrObj = instrRoot[instrumentName]

        instrObj.add_node(SpecH5LazyTiffNode(tiffPathFmt=tiffFormat,
                                             numFrames=numFrames,
                                             name="data",
                                             parent=instrObj,
                                             **(dataNodeArgs or {})))

        # create a link inside "measurement" that points to "instrument/<name>/data"
        link_target = instrObj.name.replace('measurement', 'instrument')
        measRoot.add_node(H5SoftLink(name=instrumentName,
                                     path=link_target + "/data",
                                     parent=measRoot))



#!/usr/bin/python3

from pathlib import Path
import h5py
import logging
logger = logging.getLogger(__name__)

'''
This is about writing to NX(-like) backends with limited permissions.

We're focusing on the `nx5d` data model (i.e. frame dimension first,
everything after, scan/run structure, measuring one frame at a time).
On top of that, we're implementing a `.push()` or `.append()`
functionality. We're encompassing the pushing of data points by a
`.initialize()` and a `.finalize()` call.
'''



class IngestUrlError(Exception): pass

class H5ScanSink:

    def __init__(self, source, scan, **devices):
        
        self.data_sink = source
        
        scan_path = f'{self.data_sink.group_path}/{scan}.1'
        logger.info(f'Scan {scan}: requiring path "{scan_path}" in "{self.data_sink.file_path}"')

        with h5py.File(self.data_sink.file_path, 'a') as h5:
            self.scan_path = self._ensure_nxentry(scan_path, h5).name

        logger.info(f'Scan {scan}: {self.data_sink.file_path}::{self.scan_path}')

        self._datapath = {}

        for name,dspec in devices.items():
            self.ensure_device(name, **dspec)


    def _ensure_nxentry(self, path, h5obj, attr_defaults=None):
        #
        # Recursively makes sure that `path` exists in h5obj.
        # Returns a tuple (obj, name) of the deepest element.
        # Note that if h5obj is `None`, the `obj` return value
        # will also be an invalid object handle. You can use `name`
        # to find it again once you re-open the file.
        #

        if attr_defaults is None:
            attr_defaults = { 'NX_class': 'NXentry' }
        
        i = path.find('/')
        first = path[:i] if i > 0 else path
        g = h5obj.require_group(first)

        for k,v in attr_defaults.items():
            try:
                a = g.attrs[k]
            except KeyError:
                logger.info(f'Defaulting to "{k}" -> "{v}" for "{g.name}"')
                g.attrs[k] = v

        if i > 0 and i < len(path):
            return self._ensure_nxentry(path[i+1:], h5obj=g)
        else:
            return g

        
    def _subpath(self, *pelems):
        # returns a path made up of path elements 'pelems' within
        # the root of the current scan
        return '/'.join([self.scan_path] + list(pelems))


    def ensure_device(self, name, shape, dtype, dims=None, create_args=None):
        ''' Makes sure the necessary groups and datasets exist for device.

        This creates:
        
          - A NDF5 group `instrument/{name}` if it doesn't exist
        
          - A dataset `instrument/{name}/data`, which has the shape
            `(0, <shape>)`, and is extensible in its first dimension
            (i.e. to expand on the `0`).

          - A HDF5 soft-link `measurement/{name}` pointing to
            `instrument/{name}/data`

          - If `dims` is not `None`, ... (FIXME: need to expand this
            into CDF4 / xarray territory!)

        Args:

            name: string, the name of the device
        
            shape: tuple, describing N-1 dimensions of the device. A 0-th
              dimension with length 0, editable, will be added in front.
              Generally, the shape of all dimensions except the first one
              (which is automatically added) is fixed, i.e. `maxshape`
              parameter for `.create_dataset()` is the same as the shape
              itself. However, we also accept `None` as a shape designation,
              in which case we automatically translate that to size `0`
              and `maxshape` of `None` for that particular dimension.

            dtype: Data type of the device. Since we don't have any data
              yet at this stage, the parameter is mandatory.
        
            dims: If this is different from `None` (the default), then the
              data is written in CDF4-like format, with intrinsic scaling.
              This is expected to be a `dict` with dimention names
              as its keys (strings), and the data types as the values.
              If any of the value(s) are `None`, then `dtype` is automatically
              assumed for the data type. The key `"frame"` is reserved for the
              0-th dimension and its data type is `int`, if not otherwise
              specified.
        '''

        if create_args is None:
            create_args = { 'compression': 'lzf' }

        if dims is not None:
            logger.error(f'`dims` keyword not yet implemented!')


        with h5py.File(self.data_sink.file_path, 'a',
                       **self.data_sink.h5args) as h5:
            
            dev_grp = self._ensure_nxentry(self._subpath('instrument', name), h5,
                                           attr_defaults={'NX_class': 'NXinstrument'})


            shape_len = (s if s is not None else 0 for s in shape)
            shape_max = (s for s in shape)
            data = dev_grp.create_dataset('name', dtype=dtype,
                                          shape=(0, *shape_len),
                                          maxshape=(None, *shape_max))

            self._datapath[name] = data.name
            
            meas_grp = self._ensure_nxentry(self._subpath('measurement'), h5,
                                            attr_defaults={'NX_class': 'NXcollection'})
            
            meas_grp[name] = h5py.SoftLink(data.name)


    def _append_to_device(self, name, data, h5):
        dset = h5[self._datapath[name]]
        if data.shape != dset.shape[1:]:
            raise RuntimeError(f'Data for "{name}" expected '
                               f'with shape "{dset.shape[1:]}" '
                               f'got shape {data.shape} instead.')
        
        dset.resize((dset.shape[0]+1, *dset.shape[1:]))
        
        #print("Shapes:", dset[-1].shape, data.shape)
        dset[-1] = data


    @property
    def devices(self):
        return tuple([k for k in self._datapath.keys()])


    def append(self, data=None):
        ''' Appends one single data point to the scan, across all dimensions.

        Args:

            data: the data. The preferred way is to pass a dicitonary of
              arrays in `data`, or an `xarray.Dataset`. Each dataset must
              have the shape that was initially specified with at
              the beginning (see `.ensure_device()`).
        '''
        with h5py.File(self.data_sink.file_path, 'a') as h5:
            for k in self.devices:
                self._append_to_device(k, data[k], h5)
    

class H5DataSink:
    '''
    HDF5 data sink for a single scan (run).

    Writes to HDF5 files. Tries to adhere to the ESRF-esque data
    layout. We're using "scan" and "run" pretty interchangably here,
    although they mean different things.
    '''

    def __init__(self, h5like=None, h5file=None, h5group=None, **h5args):
        ''' Initializes the sink.
        
        Args:
            h5like: File path or a string "<file>::<group>", or "<file>#<group>"

            h5file: string or Path, refers to the HDF5 file only. (FIXME: h5 object?)

            h5group: string, the group path inside the HDF5 file
        '''

        # this is a string/path, we need to open it
        if (h5like is not None) and \
           not hasattr(h5like, 'create_dataset'):
            f, g = self._url_split(h5like)
                
            self.h5like = None
            self._own_h5 = True

        # This is a HDF5 file / group that is already open
        else:
            f, g = None, None
            self.h5like = h5like
            self._own_h5 = False

            
        self.file_path = Path(f) if f is not None \
            else Path(h5like) if h5like is not None \
                 else None
        self.group_path = g or "/"

        self.h5args = h5args or {}


    def _try_split(self, text, s):
        parts = text.split(s)
        if len(parts) > 2:
            raise IngestUrlError(f'H5-like path contains multiple # or ::, don\'t know how to split')

        if len(parts) == 2:
            return parts

        return parts[0], None


    def _url_split(self, url):
        try:
            f1, g1 = self._try_split(url, '#')
        except IngestUrlError:
            f1, g1 = None, None

        try:
            f2, g2 = self._try_split(url, '::')
        except IngestUrlError:
            f2, g2 = None, None

        if g1 is not None:
            return (f1, g1)
        elif g2 is not None:
            return (f2, g2)
        elif f1 == f2:
            return f1, None

        raise IngestUrlError(f'Cannot parse url {url} -- did you have multiple # or :: in it?')


    def new_scan(self, scan, **devices):
        return H5ScanSink(self, scan, **devices)



#!/usr/bin/python3

import pytest, h5py, numpy, xarray

from nx5d.xrd import kmc3

from nx5d.xrd import signal

from pprint import pprint

# It's difficult (impossible?) to actually test whether LazyQMap
# truly delivers physically correct results. What we're focusing
# on here primarily is testing whether the interface behaves as
# expected, and data dimensionality is as we want it to be.

def data_dict(num_images=64, img_size=None, **extra_data):
    # Creates a dictionary of test data with `num_images` images.
    # If `img_size` is not defined, the size of the default kmc3
    # detector (pilatus) is used.
    #
    # `extra_data` is an additional dictionary of data, expected
    # to have the same dimensionality as the `num_images`. It
    # will be added to the dictionary.
    #
    # The dictionary structure is... biased: a main key 'angles'
    # will hold the specific angles (phi, chi, omega, twotheta).
    # The rest of the data (images, extra_data) are flat alongside
    # the angles.
    
    num_imgs = 64
    
    if img_size is None:
        img_size = kmc3.ExperimentTemplate['imageSize']
        
    return {
        'images': numpy.ndarray((num_images, *img_size)),
        'angles': {
            'phi': numpy.array([.0]*num_images),
            'chi': numpy.array([.0]*num_images),
            'theta': numpy.array(range(num_images))*0.01 + 12.0,            
            'twotheta': numpy.array(range(num_images))*0.01 + 24.0
        },
        **extra_data
    }
    
@pytest.fixture
def h5data():
    # Generating an artificial HDF5 dataset of ~50 images
    #with h5py.File() as h5:
    #    return h5
    pass

@pytest.fixture
def xrdata(data=None, index=None, **extra_data):
    # Returning an xarray.Dataset with ~50 images and corresponding angles.
    # Structurally this is different from the data_dict: _all_ data is flat,
    # including the angles.
    #
    # Returns a (dataset, exp_setup) tuple.

    if data is None:
        num_images = 64
        img_size = kmc3.ExperimentTemplate['imageSize']
        data = data_dict(num_images)

    coords = {
        'index': index if index is not None \
            else numpy.array(range(data['images'].shape[0])),
        'x': numpy.array(range(data['images'].shape[1])),
        'y': numpy.array(range(data['images'].shape[2])),
    }
            # 
    dvars = { k: ('index', v) for k,v in data['angles'].items() }
    dvars.update({ k: ('index', v) for k,v in extra_data.items() })
    dvars['images'] = ('index', 'x', 'y'), data['images']

    # Create an experimental setup suitable for this module's test data
    # (based on the KMC3 experimental setup).
    # Differently from a HDF5-based data dictionary, _all_ the data
    # is inside the `ds` Dataset here.
    ds = xarray.Dataset(coords=coords, data_vars=dvars)
    exp_setup = kmc3.ExperimentTemplate.copy()

    exp_setup.update({

        # All the angles are in the Dataset already. Here we just
        # need to name which data-vars  are to be attributed to the angles.
        #'detectorTARAngles': {
        #    'twotheta': "@twotheta"
        #},

        'goniometerAngles': {
            'chi': "@chi",
            'phi': "@phi",
            'theta': "@theta"
        },

        # Need some fake data (which won't be readable from Dataset)
        'beamEnergy': 9600.0,
        'imageDistance': 720.0,
    })

    return ds, exp_setup


def test_data(xrdata):
    print (xrdata[0])
    pprint(xrdata[1])


# FIXME: test broken
def _test_qmapper(xrdata):

    # setup in xrdata contains 'goniometerAngles' and 'detectorTARAngles'.
    clean_setup_no_angles = {
        k:xrdata[1][k] for k in filter(lambda x: not x.endswith('Angles'), xrdata[1])
    }

    clean_setup_with_angles = xrdata[1].copy()

    print("Clean setup, no angle names:")
    pprint(clean_setup_no_angles)

    # re-introduce angle names in 'Axes' parameters -> make them dict()s
    #clean_setup_with_angles = clean_setup_no_angles.copy()
    #clean_setup_with_angles['goniometerAxes'] = {
    #    k:v for k,v in zip(xrdata[1]['goniometerAngles'].keys(),
    #                       clean_setup_no_angles['goniometerAxes'])
    #}
    #clean_setup_with_angles['detectorTARAxes'] = {
    #    k:v for k,v in zip(xrdata[1]['detectorTARAngles'].keys(),
    #                       clean_setup_no_angles['detectorTARAxes'])
    #}

    print("Clean setup, with angle names:")
    pprint(clean_setup_with_angles)


    ## selecting full image
    q = signal.QMapper(**clean_setup_no_angles)
    
    #with pytest.raises(signal.InsufficientAngles):
    with pytest.raises(Exception):    
        # This is supposed to fail (default 'setup' does not include angle names)
        q.qmap(xrdata[0])
        
    q.qmap(xrdata[0], angles=('chi', 'phi', 'theta', 'twotheta'))

    ## selecting with ROI -- FIXME -- this is a clustefuck, untangling new and old LazyQMap/QMapper API
    q = signal.QMapper(**clean_setup_with_angles, roi=(0, 100, 0, 200))
    q.qmap(xrdata[0].isel(x=slice(0,100), y=slice(0,200)))

    ## plain selection, no ROI, with angles
    q = signal.QMapper(**clean_setup_with_angles)
    
    with pytest.raises(Exception):
        # must fail because default selection is 3 axes
        q.qmap(xrdata[0], qsize=(100, 100))
        
    q.qmap(xrdata[0], dims=('qx', 'qy'))
    q.qmap(xrdata[0], qsize=dict(qx=100, qy=300))
    q.qmap(xrdata[0])

    data = q.qmap(xrdata[0])

    print(data)



#!/usr/bin/python3

import nx5d.ingest as ingest
import h5py, pytest, os, tempfile, numpy
from pathlib import Path

def test_h5path():

    assert ingest.H5DataSink('/path/to/h5').file_path == Path('/path/to/h5')
    
    assert ingest.H5DataSink('/path/to/h5#group').file_path == Path('/path/to/h5')
    assert ingest.H5DataSink('/path/to/h5#group').group_path == 'group'
    
    assert ingest.H5DataSink('/path/to/h5::group').file_path == Path('/path/to/h5')
    assert ingest.H5DataSink('/path/to/h5::group').group_path == 'group'

    # This is actually not ambiguous, by definition we let '#' have
    # precedence over '::'
    assert ingest.H5DataSink('ambi::guo#us').file_path == Path('ambi::guo')
    assert ingest.H5DataSink('ambi::guo#us').group_path == 'us'
    assert ingest.H5DataSink('ambi#guo::us').file_path == Path('ambi')
    assert ingest.H5DataSink('ambi#guo::us').group_path == 'guo::us'

    with pytest.raises(ingest.IngestUrlError):
        ingest.H5DataSink('fa##il')


def test_scan():

    with tempfile.TemporaryDirectory() as tmp:
        hfile = os.path.join(tmp, "scan.h5::node")
        sink = ingest.H5DataSink(hfile)
        scan = sink.new_scan(1)


def test_device():
    with tempfile.TemporaryDirectory() as tmp:
        hfile = os.path.join(tmp, "device.h5::node")
        sink = ingest.H5DataSink(hfile)
        scan = sink.new_scan(scan=1,
                             detector={'shape': (256, 512), 'dtype': 'f'},
                             angle={'shape': tuple(), 'dtype': 'f'})

        #print(f'Devices: {scan.devices}')

        with h5py.File(sink.file_path, 'r') as h5:
            for d in scan.devices:
                dset = h5[f"/node/1.1/measurement/{d}"]
                #print(f'{d} data: {dset.shape}')
                assert dset.shape[0] == 0
            assert len(h5[f"/node/1.1/measurement/angle"].shape) == 1
            assert len(h5[f"/node/1.1/measurement/detector"].shape) == 3
        
        scan.append({
            'detector': numpy.ndarray((256, 512)),
            'angle': numpy.array(3.14)
        })

        with h5py.File(sink.file_path, 'r') as h5:
            for d in scan.devices:
                dset = h5[f"/node/1.1/measurement/{d}"]
                #print(f'{d} data: {dset.shape}')
                assert dset.shape[0] == 1

        # We only accept a full set of data with every .append() call
        with pytest.raises(KeyError):
            scan.append({
                'angle': numpy.array(6.28)
            })
